__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.types import ASGIApp, Receive, Scope, Send

from dishka import AsyncContainer, FromDishka
from .base import wrap_injection


def inject(func):
    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda r, _: r[0].scope["state"]["dishka_container"],
        additional_params=[],
        is_async=True,
    )


class ContainerMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
            self, scope: Scope, receive: Receive, send: Send,
    ) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        request = Request(scope)
        async with request.app.state.dishka_container(
                {Request: request},
        ) as request_container:
            request.state.dishka_container = request_container
            return await self.app(scope, receive, send)


def setup_dishka(container: AsyncContainer, app: Starlette) -> None:
    app.add_middleware(ContainerMiddleware)
    app.state.dishka_container = container



from collections.abc import Callable
from typing import Any, overload

from .composite import CompositeDependencySource, ensure_composite
from .decorator import Decorator
from .make_factory import make_factory
from .unpack_provides import unpack_decorator


def _decorate(
        source: Callable | type | None = None,
        provides: Any = None,
        *,
        is_in_class: bool = True,
) -> CompositeDependencySource:
    composite = ensure_composite(source)
    decorator = Decorator(make_factory(
        provides=provides, scope=None, source=source, cache=False,
        is_in_class=is_in_class,
    ))
    composite.dependency_sources.extend(unpack_decorator(decorator))
    return composite


@overload
def decorate(
        *,
        provides: Any = None,
) -> Callable[[Callable], CompositeDependencySource]:
    ...


@overload
def decorate(
        source: Callable | type,
        *,
        provides: Any = None,
) -> CompositeDependencySource:
    ...


def decorate(
        source: Callable | type | None = None,
        provides: Any = None,
) -> Any:
    if source is not None:
        return _decorate(source, provides, is_in_class=True)

    def scoped(func):
        return _decorate(func, provides, is_in_class=True)

    return scoped


def decorate_on_instance(
        source: Callable | type | None = None,
        provides: Any = None,
) -> CompositeDependencySource:
    return _decorate(source, provides, is_in_class=False)




__all__ = (
    "FromDishka",
    "inject",
    "setup_dishka",
)

from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager
from typing import Any, TypeVar

from faststream import BaseMiddleware, FastStream, context
from faststream.types import DecodedMessage

from dishka import AsyncContainer, FromDishka
from dishka.integrations.base import wrap_injection

T = TypeVar("T")


def inject(func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
    return wrap_injection(
        func=func,
        container_getter=lambda *_: context.get_local("dishka"),
        is_async=True,
        remove_depends=True,
    )


class DishkaMiddleware(BaseMiddleware):
    def __init__(self, container: AsyncContainer) -> None:
        self.container = container

    def __call__(self, msg: Any | None = None) -> "DishkaMiddleware":
        self.msg = msg
        return self

    @asynccontextmanager
    async def consume_scope(
            self,
            *args: Any,
            **kwargs: Any,
    ) -> AsyncIterator[DecodedMessage]:
        async with self.container() as request_container:
            with context.scope("dishka", request_container):
                async with super().consume_scope(*args, **kwargs) as result:
                    yield result


def setup_dishka(
        container: AsyncContainer,
        app: FastStream,
        *,
        finalize_container: bool = True,
) -> None:
    assert app.broker, "You can't patch FastStream application without broker"  # noqa: S101

    if finalize_container:
        app.after_shutdown(container.close)

    app.broker.middlewares = (
        *app.broker.middlewares,
        DishkaMiddleware(container),
    )



__all__ = [
    "AutoInjectMiddleware",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Container
from inspect import Parameter

from aiogram import BaseMiddleware, Router
from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.types import TelegramObject

from dishka import AsyncContainer, FromDishka
from .base import is_dishka_injected, wrap_injection

CONTAINER_NAME = "dishka_container"


def inject(func):
    additional_params = [Parameter(
        name=CONTAINER_NAME,
        annotation=Container,
        kind=Parameter.KEYWORD_ONLY,
    )]

    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda _, p: p[CONTAINER_NAME],
        additional_params=additional_params,
        is_async=True,
    )


class ContainerMiddleware(BaseMiddleware):
    def __init__(self, container):
        self.container = container

    async def __call__(
        self, handler, event, data,
    ):
        async with self.container({TelegramObject: event}) as sub_container:
            data[CONTAINER_NAME] = sub_container
            return await handler(event, data)


class AutoInjectMiddleware(BaseMiddleware):
    async def __call__(
        self, handler, event, data,
    ):
        old_handler: HandlerObject = data["handler"]
        if is_dishka_injected(old_handler.callback):
            return await handler(event, data)

        new_handler = HandlerObject(
            callback=inject(old_handler.callback),
            filters=old_handler.filters,
            flags=old_handler.flags,
        )
        old_handler.callback = new_handler.callback
        old_handler.params = new_handler.params
        old_handler.varkw = new_handler.varkw
        old_handler.awaitable = new_handler.awaitable
        return await handler(event, data)


def setup_dishka(
    container: AsyncContainer,
    router: Router,
    *,
    auto_inject: bool = False,
) -> None:
    middleware = ContainerMiddleware(container)
    auto_inject_middleware = AutoInjectMiddleware()

    for observer in router.observers.values():
        observer.outer_middleware(middleware)
        if auto_inject and observer.event_name != "update":
            observer.middleware(auto_inject_middleware)



from typing import TYPE_CHECKING, Any, Callable, Tuple, Type, TypeVar, Union

K_contra = TypeVar('K_contra', contravariant=True)
V_co = TypeVar('V_co', covariant=True)
T = TypeVar('T')

Loader = Callable[[Any], V_co]
Dumper = Callable[[K_contra], Any]
Converter = Callable[..., Any]
Coercer = Callable[[Any], Any]

TypeHint = Any

VarTuple = Tuple[T, ...]

Catchable = Union[Type[BaseException], VarTuple[Type[BaseException]]]

# https://github.com/python/typing/issues/684#issuecomment-548203158
if TYPE_CHECKING:
    EllipsisType = ellipsis  # pylint: disable=undefined-variable  # noqa: F821
else:
    EllipsisType = type(Ellipsis)



from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Protocol

from .dependency_source import FactoryType


@dataclass
class Exit:
    __slots__ = ("type", "callable")
    type: FactoryType
    callable: Callable


class CompiledFactory(Protocol):
    @abstractmethod
    def __call__(
            self, getter: Callable, exits: list[Exit], context: Any,
    ) -> Any:
        raise NotImplementedError



from typing import (
    Any,
)

from dishka.entities.component import DEFAULT_COMPONENT
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope
from .composite import CompositeDependencySource
from .context_var import ContextVariable, context_stub


def from_context(
        *, provides: Any, scope: BaseScope | None = None,
) -> CompositeDependencySource:
    composite = CompositeDependencySource(origin=context_stub)
    composite.dependency_sources.append(ContextVariable(
        provides=DependencyKey(
            type_hint=provides,
            component=DEFAULT_COMPONENT,
        ),
        scope=scope,
    ))
    return composite



from collections.abc import Callable
from threading import Lock
from typing import Any, Optional, TypeVar

from dishka.entities.component import DEFAULT_COMPONENT, Component
from dishka.entities.key import DependencyKey
from dishka.entities.scope import BaseScope, Scope
from .container_objects import Exit
from .dependency_source import FactoryType
from .exceptions import (
    ExitError,
    NoFactoryError,
)
from .provider import BaseProvider
from .registry import Registry, RegistryBuilder

T = TypeVar("T")


class Container:
    __slots__ = (
        "registry", "child_registries", "context", "parent_container",
        "lock", "_exits", "close_parent",
    )

    def __init__(
            self,
            registry: Registry,
            *child_registries: Registry,
            parent_container: Optional["Container"] = None,
            context: dict | None = None,
            lock_factory: Callable[[], Lock] | None = None,
            close_parent: bool = False,
    ):
        self.registry = registry
        self.child_registries = child_registries
        self.context = {DependencyKey(type(self), DEFAULT_COMPONENT): self}
        if context:
            for key, value in context.items():
                if isinstance(key, DependencyKey):
                    self.context[key] = value
                else:
                    self.context[DependencyKey(key, DEFAULT_COMPONENT)] = value
        self.parent_container = parent_container
        if lock_factory:
            self.lock = lock_factory()
        else:
            self.lock = None
        self._exits: list[Exit] = []
        self.close_parent = close_parent

    def __call__(
            self,
            context: dict | None = None,
            lock_factory: Callable[[], Lock] | None = None,
            scope: BaseScope | None = None,
    ) -> "ContextWrapper":
        """
        Prepare container for entering the inner scope.
        :param context: Data which will available in inner scope
        :param lock_factory: Callable to create lock instance or None
        :param scope: target scope or None to enter next non-skipped scope
        :return: context manager for inner scope
        """
        if not self.child_registries:
            raise ValueError("No child scopes found")
        child = Container(
            *self.child_registries,
            parent_container=self,
            context=context,
            lock_factory=lock_factory,
        )
        if scope is None:
            while child.registry.scope.skip:
                if not child.child_registries:
                    raise ValueError("No non-skipped scopes found.")
                child = Container(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        else:
            while child.registry.scope is not scope:
                if not child.child_registries:
                    raise ValueError(f"Cannot find {scope} as a child of "
                                     f"current {self.registry.scope}")
                child = Container(
                    *child.child_registries,
                    parent_container=child,
                    context=context,
                    lock_factory=lock_factory,
                    close_parent=True,
                )
        return ContextWrapper(child)

    def get(
            self,
            dependency_type: type[T],
            component: Component = DEFAULT_COMPONENT,
    ) -> T:
        lock = self.lock
        key = DependencyKey(dependency_type, component)
        if not lock:
            return self._get_unlocked(key)
        with lock:
            return self._get_unlocked(key)

    def _get_unlocked(self, key: DependencyKey) -> Any:
        if key in self.context:
            return self.context[key]
        compiled = self.registry.get_compiled(key)
        if not compiled:
            if not self.parent_container:
                raise NoFactoryError(key)
            return self.parent_container.get(
                key.type_hint, key.component,
            )
        try:
            return compiled(self._get_unlocked, self._exits, self.context)
        except NoFactoryError as e:
            e.add_path(self.registry.get_factory(key))
            raise

    def close(self) -> None:
        errors = []
        for exit_generator in self._exits[::-1]:
            try:
                if exit_generator.type is FactoryType.GENERATOR:
                    next(exit_generator.callable)
            except StopIteration:  # noqa: PERF203
                pass
            except Exception as err:  # noqa: BLE001
                errors.append(err)
        if self.close_parent:
            try:
                self.parent_container.close()
            except Exception as err:  # noqa: BLE001
                errors.append(err)

        if errors:
            raise ExitError("Cleanup context errors", errors)


class ContextWrapper:
    __slots__ = ("container",)

    def __init__(self, container: Container):
        self.container = container

    def __enter__(self) -> Container:
        return self.container

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.container.close()


def make_container(
        *providers: BaseProvider,
        scopes: type[BaseScope] = Scope,
        context: dict | None = None,
        lock_factory: Callable[[], Lock] | None = None,
        skip_validation: bool = False,
        start_scope: BaseScope | None = None,
) -> Container:
    registries = RegistryBuilder(
        scopes=scopes,
        container_type=Container,
        providers=providers,
        skip_validation=skip_validation,
    ).build()
    container = Container(
        *registries,
        context=context,
        lock_factory=lock_factory,
    )
    if start_scope is None:
        while container.registry.scope.skip:
            container = Container(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    else:
        while container.registry.scope is not start_scope:
            container = Container(
                *container.child_registries,
                parent_container=container,
                context=context,
                lock_factory=lock_factory,
                close_parent=True,
            )
    return container



__all__ = [
    "DishkaRoute",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from inspect import Parameter
from typing import Any, get_type_hints

from fastapi import FastAPI, Request
from fastapi.routing import APIRoute

from dishka import AsyncContainer, FromDishka
from .base import wrap_injection
from .starlette import ContainerMiddleware


def inject(func):
    hints = get_type_hints(func)
    request_param = next(
        (name for name, hint in hints.items() if hint is Request),
        None,
    )
    if request_param:
        additional_params = []
    else:
        request_param = "____dishka_request"
        additional_params = [Parameter(
            name=request_param,
            annotation=Request,
            kind=Parameter.KEYWORD_ONLY,
        )]

    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda _, p: p[request_param].state.dishka_container,
        additional_params=additional_params,
        is_async=True,
    )


class DishkaRoute(APIRoute):
    def __init__(
            self, path: str, endpoint: Callable[..., Any], **kwargs,
    ):
        endpoint = inject(endpoint)
        super().__init__(path, endpoint, **kwargs)


def setup_dishka(container: AsyncContainer, app: FastAPI) -> None:
    app.add_middleware(ContainerMiddleware)
    app.state.dishka_container = container



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union as AnyOf
else:
    class AnyOf:
        def __class_getitem__(cls, item):
            if isinstance(item, tuple):
                return ProvideMultiple(item)
            return item


class ProvideMultiple:
    def __init__(self, items):
        self.items = items



from collections.abc import Callable, Sequence
from typing import Any

from .alias import Alias
from .context_var import ContextVariable
from .decorator import Decorator
from .factory import Factory

DependencySource = Alias | Factory | Decorator | ContextVariable


class CompositeDependencySource:
    _instances = 0

    def __init__(
            self,
            origin: Callable,
            dependency_sources: Sequence[DependencySource] = (),
    ) -> None:
        self.dependency_sources = list(dependency_sources)
        self.origin = origin
        CompositeDependencySource._instances += 1
        self.number = self._instances

    def __get__(self, instance, owner) -> "CompositeDependencySource":
        try:
            origin = self.origin.__get__(instance, owner)
        except AttributeError:  # not a valid descriptor
            origin = self.origin
        return CompositeDependencySource(
            origin=origin,
            dependency_sources=[
                s.__get__(instance, owner) for s in self.dependency_sources
            ],
        )

    def __call__(self, *args, **kwargs) -> Any:
        return self.origin(*args, **kwargs)


def ensure_composite(
        origin: Callable | CompositeDependencySource,
) -> CompositeDependencySource:
    if isinstance(origin, CompositeDependencySource):
        return origin
    else:
        return CompositeDependencySource(origin)



__all__ = [
    "DISHKA_CONTAINER_KEY",
    "FromDishka",
    "inject",
    "setup_dishka",
]

from collections.abc import Callable
from typing import Final

from aiohttp import web
from aiohttp.typedefs import Handler
from aiohttp.web_app import Application
from aiohttp.web_request import Request
from aiohttp.web_response import StreamResponse

from dishka import AsyncContainer, FromDishka
from dishka.integrations.base import wrap_injection

DISHKA_CONTAINER_KEY: Final = web.AppKey("dishka_container", AsyncContainer)


def inject(func: Callable) -> Callable:
    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda p, _: p[0][DISHKA_CONTAINER_KEY],
        is_async=True,
    )


@web.middleware
async def container_middleware(
    request: Request, handler: Handler,
) -> StreamResponse:
    container = request.app[DISHKA_CONTAINER_KEY]
    async with container(context={Request: request}) as request_container:
        request[DISHKA_CONTAINER_KEY] = request_container
        return await handler(request)


def setup_dishka(container: AsyncContainer, app: Application) -> None:
    app[DISHKA_CONTAINER_KEY] = container
    app.middlewares.append(container_middleware)



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from flask import Flask, Request, g, request
from flask.sansio.scaffold import Scaffold

from dishka import Container, FromDishka
from .base import is_dishka_injected, wrap_injection


def inject(func):
    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda _, p: g.dishka_container,
        additional_params=[],
        is_async=False,
    )


class ContainerMiddleware:
    def __init__(self, container):
        self.container = container

    def enter_request(self):
        g.dishka_container_wrapper = self.container({Request: request})
        g.dishka_container = g.dishka_container_wrapper.__enter__()

    def exit_request(self, *_args, **_kwargs):
        g.dishka_container.close()


def _inject_routes(scaffold: Scaffold):
    for key, func in scaffold.view_functions.items():
        if not is_dishka_injected(func):
            scaffold.view_functions[key] = inject(func)


def setup_dishka(
        container: Container,
        app: Flask,
        *,
        auto_inject: bool = False,
) -> None:
    middleware = ContainerMiddleware(container)
    app.before_request(middleware.enter_request)
    app.teardown_appcontext(middleware.exit_request)
    if auto_inject:
        _inject_routes(app)
        for blueprint in app.blueprints.values():
            _inject_routes(blueprint)



import types
import typing
from typing import (
    Any,
    Dict,
    ForwardRef,
    Generic,
    Iterable,
    Protocol,
    TypedDict,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from ..common import TypeHint, VarTuple
from ..feature_requirement import HAS_ANNOTATED, HAS_PY_39, HAS_PY_312, HAS_STD_CLASSES_GENERICS
from .constants import BUILTIN_ORIGIN_TO_TYPEVARS

TYPED_DICT_MCS = type(types.new_class("_TypedDictSample", (TypedDict,), {}))


def strip_alias(type_hint: TypeHint) -> TypeHint:
    origin = get_origin(type_hint)
    return type_hint if origin is None else origin


def is_subclass_soft(cls, classinfo) -> bool:
    """Acts like builtin issubclass,
     but returns False instead of rising TypeError
    """
    try:
        return issubclass(cls, classinfo)
    except TypeError:
        return False


def has_attrs(obj, attrs: Iterable[str]) -> bool:
    return all(
        hasattr(obj, attr_name)
        for attr_name in attrs
    )


def is_new_type(tp) -> bool:
    return has_attrs(tp, ['__supertype__', '__name__'])


def is_typed_dict_class(tp) -> bool:
    return isinstance(tp, TYPED_DICT_MCS)


NAMED_TUPLE_METHODS = ('_fields', '_field_defaults', '_make', '_replace', '_asdict')


def is_named_tuple_class(tp) -> bool:
    return (
        is_subclass_soft(tp, tuple)
        and
        has_attrs(tp, NAMED_TUPLE_METHODS)
    )


def is_protocol(tp):
    if not isinstance(tp, type):
        return False

    return Protocol in tp.__bases__


def create_union(args: tuple):
    return Union[args]


if HAS_ANNOTATED:
    def get_all_type_hints(obj, globalns=None, localns=None):
        return get_type_hints(obj, globalns, localns, include_extras=True)
else:
    get_all_type_hints = get_type_hints


def is_parametrized(tp: TypeHint) -> bool:
    return bool(get_args(tp))


def get_type_vars(tp: TypeHint) -> VarTuple[TypeVar]:
    return getattr(tp, '__parameters__', ())


if HAS_PY_312:
    def is_user_defined_generic(tp: TypeHint) -> bool:
        # pylint: disable=no-member
        return (
            bool(get_type_vars(tp))
            and (
                is_subclass_soft(strip_alias(tp), Generic)
                or isinstance(tp, typing.TypeAliasType)  # type: ignore[attr-defined]
            )
        )
else:
    def is_user_defined_generic(tp: TypeHint) -> bool:
        return (
            bool(get_type_vars(tp))
            and is_subclass_soft(strip_alias(tp), Generic)
        )


def is_generic(tp: TypeHint) -> bool:
    """Check if the type could be parameterized"""
    return (
        bool(get_type_vars(tp))
        or (
            strip_alias(tp) in BUILTIN_ORIGIN_TO_TYPEVARS
            and tp != type
            and not is_parametrized(tp)
            and (
                bool(HAS_STD_CLASSES_GENERICS) or not isinstance(tp, type)
            )
        )
        or (
            bool(HAS_ANNOTATED)
            and get_origin(tp) == typing.Annotated
            and is_generic(tp.__origin__)
        )
    )


def is_bare_generic(tp: TypeHint) -> bool:
    """Check if the type could be parameterized, excluding type aliases (list[T] etc.)"""
    return (
        (
            is_generic(strip_alias(tp))
            # for 3.8 and List (list is not generic)
            or is_generic(tp)
            # at 3.8 list is bare_generic but not generic
            # (this function only needs to create predicate)
            or tp in BUILTIN_ORIGIN_TO_TYPEVARS
        )
        and not is_parametrized(tp)
    )


def is_generic_class(cls: type) -> bool:
    """Check if the class represents a generic type.
    This function is faster than ``.is_generic()``, but it is limited to testing only classes
    """
    return (
        cls in BUILTIN_ORIGIN_TO_TYPEVARS
        or (
            issubclass(cls, Generic)  # type: ignore[arg-type]
            and bool(cls.__parameters__)  # type: ignore[attr-defined]
        )
    )


def get_type_vars_of_parametrized(tp: TypeHint) -> VarTuple[TypeVar]:
    try:
        params = tp.__parameters__
    except AttributeError:
        return ()

    if isinstance(tp, type):
        if HAS_STD_CLASSES_GENERICS and isinstance(tp, types.GenericAlias):
            return params
        return ()
    if get_origin(tp) is not None and get_args(tp) == ():
        return ()
    return params


if HAS_PY_39:
    def eval_forward_ref(namespace: Dict[str, Any], forward_ref: ForwardRef):
        # pylint: disable=protected-access
        return forward_ref._evaluate(namespace, None, frozenset())
else:
    def eval_forward_ref(namespace: Dict[str, Any], forward_ref: ForwardRef):
        # pylint: disable=protected-access
        return forward_ref._evaluate(namespace, None)  # type: ignore[call-arg]



from collections.abc import Awaitable, Callable, Sequence
from inspect import Parameter, Signature, signature
from typing import (
    Annotated,
    Any,
    Literal,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from dishka.async_container import AsyncContainer
from dishka.container import Container
from dishka.entities.depends_marker import FromDishka
from dishka.entities.key import DEFAULT_COMPONENT, DependencyKey, FromComponent


def default_parse_dependency(
        parameter: Parameter,
        hint: Any,
        depends_class: type | Sequence[type] = (FromDishka, FromComponent),
) -> Any:
    """Resolve dependency type or return None if it is not a dependency."""
    if get_origin(hint) is not Annotated:
        return None
    args = get_args(hint)
    dep = next(
        (arg for arg in args if isinstance(arg, depends_class)),
        None,
    )
    if not dep:
        return None
    if isinstance(dep, (FromDishka, FromComponent)):
        return DependencyKey(args[0], dep.component)
    else:
        return DependencyKey(args[0], DEFAULT_COMPONENT)


DependencyParser = Callable[[Parameter, Any], DependencyKey | None]


@overload
def wrap_injection(
        *,
        func: Callable,
        container_getter: Callable[[tuple, dict], Container],
        is_async: Literal[False] = False,
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Callable:
    ...


@overload
def wrap_injection(
        *,
        func: Callable,
        container_getter: Callable[[tuple, dict], AsyncContainer],
        is_async: Literal[True],
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
) -> Awaitable:
    ...


def wrap_injection(
        *,
        func: Callable,
        container_getter: Callable,
        is_async: bool = False,
        remove_depends: bool = True,
        additional_params: Sequence[Parameter] = (),
        parse_dependency: DependencyParser = default_parse_dependency,
):
    hints = get_type_hints(func, include_extras=True)
    func_signature = signature(func)

    dependencies = {}
    for name, param in func_signature.parameters.items():
        hint = hints.get(name, Any)
        dep = parse_dependency(param, hint)
        if dep is None:
            continue
        dependencies[name] = dep

    if remove_depends:
        new_annotations = {
            name: hint
            for name, hint in hints.items()
            if name not in dependencies
        }
        new_params = [
            param
            for name, param in func_signature.parameters.items()
            if name not in dependencies
        ]
    else:
        new_annotations = hints.copy()
        new_params = func_signature.parameters.copy()

    if additional_params:
        new_params.extend(additional_params)
        for param in additional_params:
            new_annotations[param.name] = param.annotation

    if is_async:
        autoinjected_func = _async_injection_wrapper(
            container_getter=container_getter,
            dependencies=dependencies,
            func=func,
            additional_params=additional_params,
        )
    else:
        autoinjected_func = _sync_injection_wrapper(
            container_getter=container_getter,
            dependencies=dependencies,
            func=func,
            additional_params=additional_params,
        )

    autoinjected_func.__dishka_injected__ = True
    autoinjected_func.__name__ = func.__name__
    autoinjected_func.__qualname__ = func.__qualname__
    autoinjected_func.__doc__ = func.__doc__
    autoinjected_func.__annotations__ = new_annotations
    autoinjected_func.__signature__ = Signature(
        parameters=new_params,
        return_annotation=func_signature.return_annotation,
    )
    return autoinjected_func


def is_dishka_injected(func):
    return hasattr(func, "__dishka_injected__")


def _async_injection_wrapper(
        container_getter: Callable,
        additional_params: Sequence[Parameter],
        dependencies: dict[str, DependencyKey],
        func: Callable,
):
    async def autoinjected_func(*args, **kwargs):
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        solved = {
            name: await container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        return await func(*args, **kwargs, **solved)

    return autoinjected_func


def _sync_injection_wrapper(
        container_getter: Callable,
        additional_params: Sequence[Parameter],
        dependencies: dict[str, DependencyKey],
        func: Callable,
):
    def autoinjected_func(*args, **kwargs):
        container = container_getter(args, kwargs)
        for param in additional_params:
            kwargs.pop(param.name)
        solved = {
            name: container.get(dep.type_hint, component=dep.component)
            for name, dep in dependencies.items()
        }
        return func(*args, **kwargs, **solved)

    return autoinjected_func



__all__ = [
    "FromDishka",
    "inject",
    "setup_dishka",
]

from inspect import Parameter

import telebot
from telebot import BaseMiddleware, TeleBot

from dishka import Container, FromDishka
from .base import wrap_injection

CONTAINER_NAME = "dishka_container"


def inject(func):
    additional_params = [Parameter(
        name=CONTAINER_NAME,
        annotation=Container,
        kind=Parameter.KEYWORD_ONLY,
    )]

    return wrap_injection(
        func=func,
        remove_depends=True,
        container_getter=lambda _, p: p[CONTAINER_NAME],
        additional_params=additional_params,
        is_async=False,
    )


class ContainerMiddleware(BaseMiddleware):
    update_types = telebot.util.update_types

    def __init__(self, container):
        super().__init__()
        self.container = container

    def pre_process(self, message, data):
        dishka_container_wrapper = self.container({type(message): message})
        data[CONTAINER_NAME + "_wrapper"] = dishka_container_wrapper
        data[CONTAINER_NAME] = dishka_container_wrapper.__enter__()

    def post_process(self, message, data, exception):
        data[CONTAINER_NAME + "_wrapper"].__exit__(None, None, None)


def setup_dishka(container: Container, bot: TeleBot) -> Container:
    middleware = ContainerMiddleware(container)
    bot.setup_middleware(middleware)
    return container



import collections
import concurrent.futures
import queue
import re
from os import PathLike
from typing import Mapping, TypeVar

from ..common import VarTuple

_AnyStrT = TypeVar('_AnyStrT', str, bytes)
_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')
_T1_co = TypeVar('_T1_co', covariant=True)
_AnyStr_co = TypeVar("_AnyStr_co", str, bytes, covariant=True)

BUILTIN_ORIGIN_TO_TYPEVARS: Mapping[type, VarTuple[TypeVar]] = {
    re.Pattern: (_AnyStrT, ),  # type: ignore[dict-item]
    re.Match: (_AnyStrT, ),  # type: ignore[dict-item]
    PathLike: (_AnyStr_co, ),  # type: ignore[dict-item]
    type: (_T1,),  # type: ignore[dict-item]
    list: (_T1,),  # type: ignore[dict-item]
    set: (_T1,),  # type: ignore[dict-item]
    frozenset: (_T1_co, ),  # type: ignore[dict-item]
    collections.Counter: (_T1,),  # type: ignore[dict-item]
    collections.deque: (_T1,),  # type: ignore[dict-item]
    dict: (_T1, _T2),  # type: ignore[dict-item]
    collections.defaultdict: (_T1, _T2),  # type: ignore[dict-item]
    collections.OrderedDict: (_T1, _T2),  # type: ignore[dict-item]
    collections.ChainMap: (_T1, _T2),  # type: ignore[dict-item]
    queue.Queue: (_T1, ),  # type: ignore[dict-item]
    queue.PriorityQueue: (_T1, ),  # type: ignore[dict-item]
    queue.LifoQueue: (_T1, ),  # type: ignore[dict-item]
    queue.SimpleQueue: (_T1, ),  # type: ignore[dict-item]
    concurrent.futures.Future: (_T1, ),  # type: ignore[dict-item]
}
